/* usbtenkical: A command-line tool for USBTenki sensors.
 * Copyright (C) 2020  Dracal technologies inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <getopt.h>

#include "usbtenki.h"
#include "../common/usbtenki_cmds.h"
#include "../common/usbtenki_version.h"

#define MAX_EXTRA_ARGS 8
#define MAX_CHANNELS 32

static const char INDENT[] = "    ";
static const char NO_DEVICE_FOUND[] = "No device found.\n";
static const char UNSUPPORTED_DEVICE[] = "This device does not support user calibration.\n";

char g_verbose = 0;
char g_7bit_clean = 0;
char g_pretty = 0;

struct USBTenki_unitPreferences g_unit_prefs = {
	.temperature = TENKI_UNIT_CELSIUS,
	.pressure = TENKI_UNIT_KPA,
	.frequency = TENKI_UNIT_HZ,
	.current = TENKI_UNIT_AMPS,
	.power = TENKI_UNIT_WATTS,
	.length = TENKI_UNIT_METERS,
	.concentration = TENKI_UNIT_SENSOR_DEFAULT,
};

static void printVersion(void)
{
	printf("usbtenkical version %s\n", USBTENKI_VERSION);
	printf("Copyright (C) 2020, Dracal technologies inc.\n\n");
	printf("This software comes with ABSOLUTELY NO WARRANTY;\n");
	printf("You may redistribute copies of it under the terms of the GNU General Public License\n");
	printf("http://www.gnu.org/licenses/gpl.html\n");
}

static void printUsage(void)
{
	printf("Usage: ./usbtenkical [options] command [command arguments ...]\n");
	printf("\nValid options:\n");
	printf("    -V          Display version information\n");
	printf("    -v          Run in verbose mode\n");
	printf("    -h          Displays help\n");
	printf("    -s serial   Operate on the device matching the given serial number.\n");
	printf("    -f          Operate on the first device found.\n");
	printf("    -l          List calibration points for selected device, or all devices if none is selected.\n");
	printf("    -T unit     Select the temperature unit to use. Default: Celsius\n");
	printf("    -P unit     Select the pressure unit to use. Default: kPa\n");
	printf("    -F unit     Select the frequency unit to use. Default: Hz\n");
	printf("    -M unit     Select the length unit to use. Default: m\n");
	printf("    -C unit     Select the concentration unit to use. Default: Sensor default\n");
	printf("    -p          Enable pretty output\n");
	printf("    -7          Use 7-bit ASCII output (no Unicode degree symbols)\n");

	printf("\nTo view the list devices, channels, and points, use the '-l' option.\n");

	printf("\nTo operate on a device, select it with '-s serial' or '-f', and provide a command\n");
	printf("along with its arguments, if applicable. Each command operates on a single device.\n");

	printf("\nCommands and arguments:\n");
	printf("    set chid pid sval tval      Set a calibration point on a device.\n");
	printf("        chid    The channel ID\n");
	printf("        pid     The point ID\n");
	printf("        sval    The non-compensated source value returned by the device\n");
	printf("        tval    The compensated target value\n");
	printf("    clear chid [pid]            Clear calibration points of a single channel.\n");
	printf("        chid    The channel ID\n");
	printf("        pid     The point ID (optional; if absent, entire channel is cleared)\n");
	printf("    clearall                    Clear calibration points of all channels.\n");
	printf("        (no arguments)\n");

	printf("\nValid temperature units:\n");
	printf("    Celsius, c, Fahrenheit, f, Kelvin, k\n");
	printf("\nValid pressure units:\n");
	printf("    kPa, hPa, Pa, bar, at (98.0665 kPa), atm (101.325 kPa), Torr, psi, inHg\n");
	printf("\nValid frequency units:\n");
	printf("    mHz, Hz, kHz, MHz, rpm\n");
	printf("\nValid length units:\n");
	printf("    mm, cm, dm, m, mil, in, ft, yd\n");
	printf("\nValid concentration units:\n");
	printf("    ppb, ppm, percent\n");
}

static void printDoubleTrim(double v, const char *suffix, FILE *stream)
{
	char tmpbuf[32];
	char *p;

	snprintf(tmpbuf, sizeof(tmpbuf), "%.6f", v);

	/* If there is a decimal separator.. */
	if (strchr(tmpbuf, '.')) {
		/* as long as there are zeroes... */
		while ((p = strrchr(tmpbuf, '0')) != NULL) {
			/* in the last position (i.e. followed by a NUL) */
			if (p[1]!=0) {
				break;
			}
			*p = 0; /* remove them! */
		}
	}

	/* The above results in . being the last character when there
	 * was only zeros! (for instance, snprintf wrote 1.000000). Remove
	 * the . in that case. */
	p = strrchr(tmpbuf, '.');
	if (p && p[1]==0) {
		*p = 0;
	}

	fputs(tmpbuf, stream);
	if (suffix) {
		fputs(suffix, stream);
	}
}

static void print_point(int channel_id, int point_id, struct USBTenki_calpoint *pt, int unit)
{
	if (g_pretty) {
		printf("%s%sPoint %d: ", INDENT, INDENT, point_id);
	}
	else {
		printf("%d, %d, ", channel_id, point_id);
	}

	if (usbtenki_isCalpointSet(pt)) {
		printDoubleTrim(pt->sensed_value, ", ", stdout);
		printDoubleTrim(pt->real_value, ", ", stdout);
		printf(" [%s]\n", unitToString(unit, g_7bit_clean));

	}
	else {
		printf("** none **\n");
	}
}

static int print_channel(USBTenki_dev_handle hdl, struct USBTenki_info *info, struct USBTenki_channel *channels, int channel_id)
{
	struct USBTenki_calpoint points[USBTENKI_MAX_CALPOINTS];
	int num_points = usbtenki_getChannelCalibrationPoints(hdl, channel_id, points);

	if (num_points == 0) {
		if (!g_pretty) {
			printf("%s, ", info->str_serial);
		}
		printf("%s%s", (g_pretty ? INDENT : ""), UNSUPPORTED_DEVICE);
		return 1;
	}

	if (g_pretty) {
		printf("%sChannel %d: %s [%s]\n",
				INDENT,
				channel_id,
				chipToString(channels[channel_id].chip_id),
				chipToShortString(channels[channel_id].chip_id));
	}

	// Determine the native unit & preferred unit for this channel
	int chip_unit = usbtenki_getChipUnit(channels[channel_id].chip_id);
	int pref_unit = usbtenki_getPreferredUnit(chip_unit, &g_unit_prefs);
	if (pref_unit == TENKI_UNIT_SENSOR_DEFAULT) {
		pref_unit = chip_unit;
	}

	// Convert calibration points to the preferred unit
	usbtenki_convertCalpointsUnits(points, num_points, chip_unit, pref_unit);

	for (int point_id = 0; point_id < num_points; point_id++) {
		if (!g_pretty) {
			printf("%s, ", info->str_serial);
		}
		print_point(channel_id, point_id, &points[point_id], pref_unit);
	}

	return 0;
}

static void print_device_info(struct USBTenki_info *info, int num_channels)
{

	printf("Device: '%s', Serial: '%s', Version %d.%d, Channels: %d\n",
			info->str_prodname,
			info->str_serial,
			info->major,
			info->minor,
			num_channels);

}

static int cmd_show(int argc, char **argv, struct USBTenki_info *info, USBTenki_dev_handle hdl)
{
	struct USBTenki_channel channels[MAX_CHANNELS];
	int num_channels = usbtenki_listChannels(hdl, channels, sizeof(channels));

	if (g_pretty) {
		print_device_info(info, num_channels);
	}

	for (int channel_id = 0; channel_id < num_channels; channel_id++) {
		if (print_channel(hdl, info, channels, channel_id)) {
			return 1;
		}
	}

	return 0;
}

static int cmd_set(int argc, char **argv, struct USBTenki_info *info, USBTenki_dev_handle hdl)
{
	int num_channels, channel_id, point_id;
	struct USBTenki_calpoint calpoint;
	char *e;
	int ret;
	struct USBTenki_channel channels[MAX_CHANNELS];

	// set_calpoint channel point_id sensed_value real_value
	if (argc != 5) {
		fprintf(stderr, "Unexpected number of arguments\n");
		return -1;
	}

	channel_id = strtol(argv[1], &e, 0);
	if (e == argv[1]) {
		fprintf(stderr, "Invalid channel number\n");
		return -1;
	}

	point_id = strtol(argv[2], &e, 0);
	if (e == argv[2]) {
		fprintf(stderr, "Invalid point ID\n");
		return -1;
	}

	calpoint.sensed_value = strtod(argv[3], &e);
	if (e == argv[3]) {
		fprintf(stderr, "Invalid source value\n");
		return -1;
	}

	calpoint.real_value = strtod(argv[4], &e);
	if (e == argv[4]) {
		fprintf(stderr, "Invalid target value\n");
		return -1;
	}

	num_channels = usbtenki_listChannels(hdl, channels, sizeof(channels));
	if (channel_id >= num_channels) {
		fprintf(stderr, "Channel %d does not exist\n", channel_id);
		return -1;
	}

	// Determine the native unit & preferred unit for this channel
	int chip_unit = usbtenki_getChipUnit(channels[channel_id].chip_id);
	int pref_unit = usbtenki_getPreferredUnit(chip_unit, &g_unit_prefs);
	if (pref_unit == TENKI_UNIT_SENSOR_DEFAULT) {
		pref_unit = chip_unit;
	}

	// Convert calibration points to the chip unit
	usbtenki_convertCalpointsUnits(&calpoint, 1, pref_unit, chip_unit);

	ret = usbtenki_setUserCalibration(hdl, channel_id, point_id, &calpoint);

	if (ret) {
		fprintf(stderr, "Error (%d)\n", ret);
	}
	else {
		print_device_info(info, num_channels);
		print_channel(hdl, info, channels, channel_id);
		printf("Successfully set calibration point %d of channel %d\n", point_id, channel_id);
	}

	return ret;
}

static int cmd_clear(int argc, char **argv, struct USBTenki_info *info, USBTenki_dev_handle hdl)
{
	int channel_id;
	int point_id = -1;
	int ret = 0;
	char *e;

	int num_channels;
	int num_points;
	struct USBTenki_channel channels[MAX_CHANNELS];
	struct USBTenki_calpoint points[USBTENKI_MAX_CALPOINTS];

	struct USBTenki_calpoint calpoint = {
		NAN, NAN,
	};

	// clear channel [point_id]
	if (argc != 2 && argc != 3) {
		fprintf(stderr, "Unexpected number of arguments\n");
		return -1;
	}

	channel_id = strtol(argv[1], &e, 0);
	if (e == argv[1]) {
		fprintf(stderr, "Invalid channel number\n");
		return -1;
	}

	if (argc >= 3) {
		point_id = strtol(argv[2], &e, 0);
		if (e == argv[2] || point_id < 0) {
			fprintf(stderr, "Invalid point ID\n");
			return -1;
		}
	}

	num_channels = usbtenki_listChannels(hdl, channels, sizeof(channels));
	if (channel_id >= num_channels) {
		fprintf(stderr, "Channel %d does not exist\n", channel_id);
		return -1;
	}

	num_points = usbtenki_getChannelCalibrationPoints(hdl, channel_id, points);
	if (!num_points) {
		fprintf(stderr, UNSUPPORTED_DEVICE);
		return -1;
	}

	if (point_id >= 0) {
		if (point_id < num_points) {
			ret = usbtenki_setUserCalibration(hdl, channel_id, point_id, &calpoint);
		}
		else {
			fprintf(stderr, "Invalid point ID\n");
			return -1;
		}
	}
	else {
		// Clear all points on this channel
		for (int i = 0; i < num_points; i++) {
			ret = usbtenki_setUserCalibration(hdl, channel_id, i, &calpoint);
			if (ret) {
				break;
			}
		}
	}

	if (ret) {
		fprintf(stderr, "Error (%d)\n", ret);
	}
	else {
		print_device_info(info, num_channels);
		print_channel(hdl, info, channels, channel_id);
		if (point_id >= 0) {
			printf("Successfully cleared calibration point %d of channel %d\n", point_id, channel_id);
		}
		else {
			printf("Successfully cleared calibration points of channel %d\n", channel_id);
		}
	}

	return ret;
}

static int cmd_clearall(int argc, char **argv, struct USBTenki_info *info, USBTenki_dev_handle hdl)
{
	int num_channels, chn, num_points, pt;
	int ret = 0;
	struct USBTenki_calpoint points[USBTENKI_MAX_CALPOINTS];
	struct USBTenki_calpoint calpoint = {
		NAN, NAN,
	};

	num_channels = 	usbtenki_getNumChannels(hdl);
	for (chn = 0; chn < num_channels; chn++) {
		num_points = usbtenki_getChannelCalibrationPoints(hdl, chn, points);
		if (!num_points) {
			fprintf(stderr, UNSUPPORTED_DEVICE);
			return -1;
		}
		for (pt = 0; pt < num_points; pt++) {
			ret = usbtenki_setUserCalibration(hdl, chn, pt, &calpoint);
			if (ret) {
				fprintf(stderr, "Error (%d)\n", ret);
				return ret;
			}
		}
	}

	print_device_info(info, num_channels);
	printf("Successfully cleared all calibration points\n");

	return 0;

}

static int list_all() {

	usbtenki_init();

	struct usb_device *dev;
	struct USBTenki_info info;
	struct USBTenki_list_ctx *listContext = usbtenki_allocListCtx();
	char device_found = 0;

	while ((dev = usbtenki_listDevices(&info, listContext))) {

		USBTenki_dev_handle hdl = usbtenki_openDevice(dev);

		if (!hdl) {
			continue;
		}

		device_found |= 1;
		cmd_show(0, NULL, &info, hdl);
		usbtenki_closeDevice(hdl);

	}

	usbtenki_freeListCtx(listContext);

	if (!device_found) {
		printf(NO_DEVICE_FOUND);
		return 1;
	}

	return 0;

}

typedef int (*command_handler_t)(int argc, char **argv, struct USBTenki_info *info, USBTenki_dev_handle hdl);

static struct {
		const char *command;
		command_handler_t handler;
	} handlers[] = {
		{ "set", cmd_set },
		{ "clear", cmd_clear },
		{ "clearall", cmd_clearall },
		{ NULL, NULL }, // terminator
};

int main(int argc, char **argv)
{
	int res, i;
	int use_first = 0;
	int list = 0;
	char *use_serial = NULL;
	int n_extra_args=0;
	char *eargv[MAX_EXTRA_ARGS];
	USBTenki_dev_handle hdl = NULL;
	struct usb_device *cur_dev, *dev=NULL;
	struct USBTenki_list_ctx *listContext;
	struct USBTenki_info info;
	int retval = 0;


	// Set this so processing stops as soon as a nonoption argument is encountered. Otherwise
	// negative values are treated as arguments...
#ifdef _WIN32
	_putenv("POSIXLY_CORRECT=1");
#else
	setenv("POSIXLY_CORRECT", "1", 0);
#endif

	while (-1 != (res=getopt(argc, argv, "Vvhlfs:p7T:P:F:M:C:")))
	{
		switch (res)
		{
			case 'l':
				list = 1;
				break;

			case 'f':
				use_first = 1;
				break;

			case 'v':
				g_verbose = 1;
				break;

			case 'V':
				printVersion();
				return 0;

			case 'h':
				printUsage();
				return 0;

			case 's':
				use_serial = optarg;
				break;

			case 'T':
				if (strcasecmp(optarg, "Celsius")==0 ||
						strcasecmp(optarg, "Celcius")==0 ||  // backward-compatibility with old typo
						strcasecmp(optarg, "C")==0)
					g_unit_prefs.temperature = TENKI_UNIT_CELSIUS;
				else if (strcasecmp(optarg, "Fahrenheit")==0 ||
						strcasecmp(optarg, "F")==0)
					g_unit_prefs.temperature = TENKI_UNIT_FAHRENHEIT;
				else if (strcasecmp(optarg, "Kelvin")==0 ||
						strcasecmp(optarg, "K")==0)
					g_unit_prefs.temperature = TENKI_UNIT_KELVIN;
				else {
					fprintf(stderr, "Unknown temperature format: '%s'\n",
										optarg);
					return -1;
				}
				break;

			case 'F':
				{
					struct {
						const char *name;
						int fmt;
					} tbl[] = {
						{ "mHz", TENKI_UNIT_MILLIHZ },
						{ "Hz", TENKI_UNIT_HZ },
						{ "kHz", TENKI_UNIT_KHZ },
						{ "MHz", TENKI_UNIT_MHZ },
						{ "rpm", TENKI_UNIT_RPM },
					};

					for (i=0; i<ARRAY_SIZE(tbl); i++) {
						if (strcasecmp(tbl[i].name, optarg)==0) {
							g_unit_prefs.frequency = tbl[i].fmt;
							break;
						}
					}
					if (i==ARRAY_SIZE(tbl)) {
						fprintf(stderr,
							"Unknown frequency unit: '%s'\n", optarg);
					}
				}
				break;

			case 'P':
				{
					printf("Parsing P option\n");
					struct {
						const char *name;
						int fmt;
					} tbl[] = {
						{ "pa", TENKI_UNIT_PA },
						{ "kpa", TENKI_UNIT_KPA },
						{ "hpa", TENKI_UNIT_HPA },
						{ "bar", TENKI_UNIT_BAR },
						{ "at", TENKI_UNIT_AT },
						{ "atm", TENKI_UNIT_ATM },
						{ "torr", TENKI_UNIT_TORR },
						{ "psi", TENKI_UNIT_PSI },
						{ "inhg", TENKI_UNIT_INHG },
					};

					for (i=0; i<ARRAY_SIZE(tbl); i++) {
						if (strcasecmp(tbl[i].name, optarg)==0) {
							g_unit_prefs.pressure = tbl[i].fmt;
							printf("Using pressure unit %s\n", unitToString(g_unit_prefs.pressure, 0));
							break;
						}
					}
					if (i==ARRAY_SIZE(tbl)) {
						fprintf(stderr,
							"Unknown pressure format: '%s'\n", optarg);
					}
				}
				break;

			case 'M':
				{
					struct {
						const char *name;
						int fmt;
					} tbl[] = {
						{ "m", TENKI_UNIT_METERS },
						{ "dm", TENKI_UNIT_DECIMETERS },
						{ "cm", TENKI_UNIT_CENTIMETERS },
						{ "mm", TENKI_UNIT_MILLIMETERS },
						{ "mil", TENKI_UNIT_MILS },
						{ "mils", TENKI_UNIT_MILS },
						{ "in", TENKI_UNIT_INCHES },
						{ "ft", TENKI_UNIT_FEET },
						{ "yd", TENKI_UNIT_YARDS },
					};

					for (i=0; i<ARRAY_SIZE(tbl); i++) {
						if (strcasecmp(tbl[i].name, optarg)==0) {
							g_unit_prefs.length = tbl[i].fmt;
							break;
						}
					}
					if (i==ARRAY_SIZE(tbl)) {
						fprintf(stderr,
							"Unknown length unit: '%s'\n", optarg);
					}
				}
				break;

			case 'C':
				{
					struct {
						const char *name;
						int fmt;
					} tbl[] = {
						{ "ppb", TENKI_UNIT_PPB },
						{ "ppm", TENKI_UNIT_PPM },
						{ "percent", TENKI_UNIT_PERCENT },
						{ "sensor", TENKI_UNIT_SENSOR_DEFAULT },
					};

					for (i=0; i<ARRAY_SIZE(tbl); i++) {
						if (strcasecmp(tbl[i].name, optarg)==0) {
							g_unit_prefs.concentration = tbl[i].fmt;
							break;
						}
					}
					if (i==ARRAY_SIZE(tbl)) {
						fprintf(stderr,
							"Unknown concentration unit: '%s'\n", optarg);
					}
				}
				break;

			case 'p':
				g_pretty = 1;
				break;

			case '7':
				g_7bit_clean = 1;
				break;

		}
	}

	n_extra_args = argc - optind;

	if (g_verbose) {
		printf("Extra args: %d\n", n_extra_args);
	}

	for (i=optind; i<argc; i++) {
		eargv[i-optind] = argv[i];
		if (g_verbose)
			printf("  %d: %s\n", i-optind, eargv[i-optind]);
	}

	// Validate actions (-l or command) and execute list_all() if nothing is specified

	char device_specified = (use_serial || use_first);

	if (list && n_extra_args) {
		// Both -l and command are specified; error
		fprintf(stderr, "Must specify either -l or a command, not both.\n");
		return 1;
	}
	else if (list) {
		// Only -l is specified
		if (!device_specified) {
			list_all();
			return 0;
		}
	}
	else if (n_extra_args) {
		// Only a command is specified
		if (!device_specified) {
			fprintf(stderr, "Serial number is required.\n");
			return 1;
		}
	}
	else {
		// No action is specified
		if (device_specified) {
			fprintf(stderr, "Missing command or -l option.\n");
			return 1;
		}
		else {
			// no action and no device: just list all
			list_all();
			return 0;
		}
	}

	// Execute action

	usbtenki_init();

	listContext = usbtenki_allocListCtx();

	while ((cur_dev = usbtenki_listDevices(&info, listContext))) {
		if (use_serial) {
			if (strcmp(use_serial, info.str_serial)==0) {
				dev = cur_dev;
				break;
			}
		}
		else if (use_first) {
			dev = cur_dev;
			break;
		}
	}

	usbtenki_freeListCtx(listContext);

	if (!dev) {
		if (use_serial) {
			fprintf(stderr, "Could not locate device with serial '%s'. Try usbtenkical -l\n", use_serial);
		}
		else {
			fprintf(stderr, NO_DEVICE_FOUND);
		}
		return 1;
	}

	hdl = usbtenki_openDevice(dev);
	if (!hdl) {
		fprintf(stderr, "Cannot open device\n");
		return 2;
	}

	if (list) {
		retval = cmd_show(0, NULL, &info, hdl);
		usbtenki_closeDevice(hdl);
		return retval;
	}

	for (i=0; handlers[i].command; i++) {
		if (strcmp(eargv[0], handlers[i].command)==0) {
			retval = handlers[i].handler(n_extra_args, eargv, &info, hdl);
			usbtenki_closeDevice(hdl);
			return retval;
		}
	}

	fprintf(stderr, "Unknown command '%s'\n", eargv[0]);

	return 1;

}
