#include "CalibrationEditDialog.h"
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QGridLayout>
#include <QDoubleSpinBox>
#include <QDebug>
#include <math.h>
#include "globals.h"
#include "usbtenki.h"
#include "ConfigCheckbox.h"
#include "EditButton.h"

CalibrationEditDialog::CalibrationEditDialog(QWidget *parent, TenkiDevice *td, int channel_id) : QDialog(parent)
{
	USBTenki_channel *chn;
	QDoubleValidator *value_validator;
	int pref_unit;

	chnid = channel_id;
	tenki_device = td;

	chn = tenki_device->getChannelData(channel_id);

	btn_ok = new QPushButton(tr("Apply"));
	btn_ok->setDefault(true);

	btn_cancel = new QPushButton(tr("Close"));

	connect(btn_ok, SIGNAL(clicked()), this, SLOT(applyClose()));
	connect(btn_cancel, SIGNAL(clicked()), this, SLOT(close()));

	//
	// +------------------------------------------+
	// |           Sensed    Calibrated           |  0
	// | Point 1  [     ] unit  [      ] unit   (X)      |  1
	// | Point 2  [     ] unit  [      ] unit   (X)      |  2
	// | Point 3  [     ] unti  [      ] unit   (X)      |  3
	// |                                          |  4
	// | [ ] Disable calibration for this channel |
	// |                                          |
	// |                                          |
	// |                              OK  CANCEL  |
	// +------------------------------------------|
	//
	// structure:
	//
	// VBoxLayout
	//   GridLayout
	//      points, labels, line editors, clear buttons...
	//   Spacing
	//   Checkbox
	//   Spacing
	//     -- stretch --
	//   VHBoxLayout
	//      Ok button, cancel button

	QGridLayout *lay_points = new QGridLayout();

	lbl_sensed = new QLabel("Sensed");
	lbl_calibrated = new QLabel("Calibrated");

	lay_points->addWidget(lbl_sensed, 0, 1);
	lay_points->addWidget(lbl_calibrated, 0, 3);
	value_validator = new QDoubleValidator(this);
	for (int i=0; i<chn->num_calpoints; i++) {

		QLabel *lbl_pt = new QLabel("Point " + QString::number(i+1));
		QString val_sensed = "", val_calibrated = "";
		struct USBTenki_calpoint calpoint;

		// Convert native values to display (pref) values
		int chip_unit = usbtenki_getChipUnit(chn->chip_id);
		pref_unit = usbtenki_getPreferredUnit(chip_unit, &g_tenkisources->units);
		if (pref_unit == TENKI_UNIT_SENSOR_DEFAULT) {
			pref_unit = chip_unit;
		}
		memcpy(&calpoint, &chn->calpoints[i], sizeof(struct USBTenki_calpoint));
		usbtenki_convertCalpointsUnits(&calpoint, 1, chip_unit, pref_unit);

		// If this point is valid, pre-populate the fileds with its values.
		if (usbtenki_isCalpointSet(&chn->calpoints[i])) {
			g_tenkisources->toTrimmedDoubleString(&val_sensed, calpoint.sensed_value);
			g_tenkisources->toTrimmedDoubleString(&val_calibrated, calpoint.real_value);
		}

		QLineEdit *le_sensed = new QLineEdit(val_sensed);
		le_sensed->setValidator(value_validator);
		lst_le_sensed.append(le_sensed);

		QLineEdit *le_calibrated = new QLineEdit(val_calibrated);
		le_calibrated->setValidator(value_validator);
		lst_le_calibrated.append(le_calibrated);

		EditButton *btn_clear = new EditButton("Delete", "Delete this point", ":edit-delete.png", i);
		connect(btn_clear, SIGNAL(buttonIdClicked(int)), this, SLOT(clearPointClicked(int)));

		QLabel *lbl_unit1 = new QLabel(QString::fromUtf8(unitToString(pref_unit, 0)));
		QLabel *lbl_unit2 = new QLabel(QString::fromUtf8(unitToString(pref_unit, 0)));

		lay_points->addWidget(lbl_pt, 1+i, 0);
		lay_points->addWidget(le_sensed, 1+i, 1);
		lay_points->addWidget(lbl_unit1, 1+i, 2);
		lay_points->addWidget(le_calibrated, 1+i, 3);
		lay_points->addWidget(lbl_unit2, 1+i, 4);
		lay_points->setColumnMinimumWidth(5, 30);
		lay_points->addWidget(btn_clear, 1+i, 6);
	}

	QString f;
	f.sprintf("%s:%02d", td->getSerial(), channel_id);
	cb_disable = new ConfigCheckbox(
										"Disable calibration for this channel",
										"disableCalibration/" + f);
	calibration_disabled = cb_disable->isChecked();


	QHBoxLayout *lay_btns = new QHBoxLayout();
	lay_btns->addStretch();
	lay_btns->addWidget(btn_ok);
	lay_btns->addWidget(btn_cancel);

	QVBoxLayout *lay_vert = new QVBoxLayout();
	lay_vert->addLayout(lay_points);
	lay_vert->addSpacing(10);
	lay_vert->addWidget(cb_disable);
	lay_vert->addSpacing(10);
	lay_vert->addStretch();
	lay_vert->addLayout(lay_btns);

	setLayout(lay_vert);

	setWindowTitle(tr("Calibrate channel"));

	qDebug() << "Calibration dialog for Channel id " << chnid;
}

CalibrationEditDialog::~CalibrationEditDialog()
{
//	qDebug() << "Destroying dialog";
}

double CalibrationEditDialog::textToCalpointValue(QString text)
{
	QString t = text.trimmed();


	if (t.isEmpty() || t.isNull()) {
//		qDebug() << "Text '" + text + "' treated as NAN";
		return NAN;
	}

//	qDebug() << "Text '" + text + "' treated as double";
	return text.toDouble();

}

void CalibrationEditDialog::applyClose()
{
	int i;
	struct USBTenki_calpoint calpoints[USBTENKI_MAX_CALPOINTS];
	int num_calpoints = 0, res;
	USBTenki_channel *chn;
	chn = tenki_device->getChannelData(chnid);

	qDebug() << "Apply close";

	// Convert values in widgets to calpoints
	for (i=0; i<lst_le_sensed.size(); i++) {
		calpoints[i].sensed_value = textToCalpointValue(lst_le_sensed.at(i)->text());
		calpoints[i].real_value = textToCalpointValue(lst_le_calibrated.at(i)->text());
		num_calpoints++;
	}

	// Convert display (preference) values back to the native value
	int chip_unit = usbtenki_getChipUnit(chn->chip_id);
	int pref_unit = usbtenki_getPreferredUnit(chip_unit, &g_tenkisources->units);
	if (pref_unit == TENKI_UNIT_SENSOR_DEFAULT) {
		pref_unit = chip_unit;
	}
	usbtenki_convertCalpointsUnits(calpoints, num_calpoints, pref_unit, chip_unit);

	// Sync calpoints
	for (i=0; i<lst_le_sensed.size(); i++) {
		res = tenki_device->setUserCalibration(chnid, i, &calpoints[i]);
		if (res) {
			qDebug() << "Error saving calpoint " + QString::number(i);
		}
	}

	// Copy the potentially changed checkbox state to calibration_disabled.
	// The creator of this dialog box will then call calibrationDisabled()
	// to check its status.
	calibration_disabled = cb_disable->isChecked();

	close();
}

void CalibrationEditDialog::close()
{
	QDialog::close();
}

bool CalibrationEditDialog::calibrationDisabled()
{
	return calibration_disabled;
}

void CalibrationEditDialog::clearPointClicked(int id)
{
	if (id > lst_le_sensed.size()) {
		return ;
	}

	lst_le_sensed.at(id)->setText("");
	lst_le_calibrated.at(id)->setText("");
}

