#include <stdint.h>
#include <QDebug>
#include <QSettings>
#include <QMessageBox>
#include <QLineEdit>
#include <QToolButton>
#include "DashSensorDevice.h"
#include "TenkiDevice.h"
#include "SourceAliasEdit.h"
#include "ConfigCheckbox.h"
#include "globals.h"
#include "usbtenki_cmds.h"
#include "CalibrationEditDialog.h"

#define LBL_CAL_SUSPENDED "<img src=':cal_suspended.png'>"
#define LBL_CAL_ENABLED "<img src=':cal_active.png'>"
#define TOOLTIP_CAL_SUSPENDED "Warning: Calibration points are present but calibration is disabled"
#define TOOLTIP_CAL_ACTIVE "Calibration points present and calibration active"

DashSensorDevice::DashSensorDevice(TenkiDevice *td)
{
	int col=0, i, graph_col, bigview_col;

	// Compute base title in format "SERIAL : Product".
	// This title is reused in in ::refresh() to append
	// an error message.
	base_title = QString::fromLocal8Bit(td->getSerial());
	base_title += " : " + QString(td->getProductName());
	setTitle(base_title);

	setObjectName("source"); // selector for stylesheet
	tenki_device = td;

	layout = new QGridLayout();
	setLayout(layout);

	layout->setVerticalSpacing(1);
	layout->setHorizontalSpacing(10);

//	layout->addWidget(new QLabel("<b>Channel</b>"), 0, col++);
	layout->addWidget(new QLabel("<b>Source ID</b>"), 0, col++);
	layout->setColumnMinimumWidth(col, 4);
	layout->addWidget(new QLabel("<b>Description</b>"), 0, col++);
	layout->setColumnMinimumWidth(col, 4);
	layout->addWidget(new QLabel("<b>Type</b>"), 0, col++);
	layout->setColumnMinimumWidth(col, 4);
	layout->addWidget(new QLabel("<b>Current</b>"), 0, col++);
	layout->setColumnMinimumWidth(col, 4);
	layout->addWidget(new QLabel("<b>Min.</b>"), 0, col++);
	layout->setColumnMinimumWidth(col, 4);
	layout->addWidget(new QLabel("<b>Max.</b>"), 0, col++);
	layout->setColumnMinimumWidth(col, 4);
	layout->addWidget(new QLabel("<b></b>"), 0, col++);
	layout->setColumnStretch(col, 0);
	layout->addWidget(new QLabel("<b>Unit</b>"), 0, col++);
	layout->setColumnMinimumWidth(col, 4);
	layout->addWidget(new QLabel("<b>Alias</b>"), 0, col++);
	layout->setColumnStretch(col, 0);
	bigview_col = col;
	layout->addWidget(new QLabel("<b>Big View</b>"), 0, col++);
	layout->setColumnStretch(col, 0);
	graph_col = col;
	layout->addWidget(new QLabel("<b>Graph</b>"), 0, col++);
	layout->setColumnMinimumWidth(col, 4);

	layout->addWidget(new QLabel(""), 0, col++);
	layout->setColumnMinimumWidth(col, 0);

	layout->addWidget(new QLabel(""), 0, col++);
	layout->setColumnMinimumWidth(col, 16);

	for (i=0; i<tenki_device->getNumChannels(); i++)
	{
		if (tenki_device->isChannelHidden(i)) {
			continue;
		}

		addChannel(i, i+1);
	}

	infobtn = new QPushButton(QIcon(":help-about.png"), tr("About"));
	connect(infobtn, SIGNAL(clicked(bool)), this, SLOT(infoClicked(bool)));

	cb_all_graph = new QCheckBox();
	cb_all_graph->setToolTip(tr("Select All / None"));
	connect(cb_all_graph, SIGNAL(stateChanged(int)), this, SLOT(cbAllGraphChanged(int)));

	cb_all_bigview = new QCheckBox();
	cb_all_bigview->setToolTip(tr("Select All / None"));
	connect(cb_all_bigview, SIGNAL(stateChanged(int)), this, SLOT(cbAllBigviewChanged(int)));

	layout->addWidget(infobtn, i + 1, 0);

	layout->addWidget(cb_all_graph, i + 1, graph_col);
	layout->addWidget(cb_all_bigview, i + 1, bigview_col);

	setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Expanding);
}

void DashSensorDevice::addChannel(int chn, int row)
{
	QSettings settings;
	USBTenki_channel ch;
	QString a, b, c, d, e, f, g;
	QString visible_name;
	QLabel *value_label, *unit_label;
	QLabel *tmp_label;
	MinMaxResettable *min, *max;
	QToolButton *rst;
	EditButton *calbtn;
	int col=0;

	g_tenkisources->convertToUnits(tenki_device->getChannelData(chn), &ch);

/*	
	a.sprintf("%d", chn);
	layout->addWidget(new QLabel(a), row, col++);*/

	f.sprintf("%s:%02d", tenki_device->getSerial(), chn);

	//visible_name.sprintf("[%s:%02d]", tenki_device->getSerial(), chn);
	visible_name.sprintf("%s:%02d", tenki_device->getSerial(), chn);
	tmp_label = new QLabel(visible_name);
	tmp_label->setTextInteractionFlags(Qt::TextSelectableByMouse);
	layout->addWidget(tmp_label, row, col++);

	b = QString::fromLocal8Bit(chipToString(ch.chip_id));

	tmp_label = new QLabel(b);
	descriptions.append(tmp_label);

	chip_ids.append(ch.chip_id);

//	tmp_label->setWordWrap(true);
//	tmp_label->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
	layout->addWidget(tmp_label, row, col++);

	c = QString::fromLocal8Bit(chipToShortString(ch.chip_id));
	layout->addWidget(new QLabel(c), row, col++);

	// Current value
	g_tenkisources->formatValue(&d, ch.converted_data);
	value_label = new QLabel(d);
	values.append(value_label);
	layout->addWidget(value_label, row, col++);

	// Minimum value
	min = new MinMaxResettable(true); // Minimum tracking mode
	minimums.append(min);
	layout->addWidget(min, row, col++);

	// Maximum value
	max = new MinMaxResettable(false); // Maximum tracking mode
	maximums.append(max);
	layout->addWidget(max, row, col++);

	// Min/Max reset button
	rst = new QToolButton();
	rst->setText("reset");
	rst->setIcon(QIcon(":undo.png"));
	rst->setToolTip("Reset min/max");
	rst->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
	layout->addWidget(rst, row, col++);
	QObject::connect(rst, SIGNAL(clicked()), min, SLOT(reset()));
	QObject::connect(rst, SIGNAL(clicked()), max, SLOT(reset()));

	e = QString::fromUtf8(unitToString(ch.converted_unit, 0));
	unit_label = new QLabel(e);
	layout->addWidget(unit_label, row, col++);
	units.append(unit_label);

	// alias
	SourceAliasEdit *se = new SourceAliasEdit(f);
	layout->addWidget(se, row, col++);
	connect(se, SIGNAL(sourceAliasChanged(QString,QString)), g_tenkisources, SLOT(updateAlias(QString,QString)));

	// In bigview
	ConfigCheckbox *ccb = new ConfigCheckbox("", "bigviewChecked/" + f);
	layout->addWidget(ccb, row, col++);
	bigviewCheckboxes.append(ccb);

	// In graph
	ConfigCheckbox *ccb_gr = new ConfigCheckbox("", "graphChecked/" + f);
	graphCheckboxes.append(ccb_gr);
	layout->addWidget(ccb_gr, row, col++);

	// No calibration on virtual channels. (>=256)
	if (ch.channel_id < 256) {
		// "Calibrate" button
		calbtn = new EditButton("Calibrate", "Edit calibration for this channel", chn);
		calbtn->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);
		if (ch.num_calpoints <= 0) {
			calbtn->setDisabled(true);
			calbtn->setToolTip("Calibration is not supported by this device.");
		}
		if (!settings.value("ui/show_cal_buttons").toBool()) {
			// Note: Visibility is also changed in syncAttentionIcons
			calbtn->setVisible(false);
		}

		layout->addWidget(calbtn, row, col++);
		connect(calbtn, SIGNAL(buttonIdClicked(int)), this, SLOT(calibrateClicked(int)));
		calButtons.append(calbtn);

		// Attention icon shown when user-calibration is disabled or when it is active.
		// The tool top and text/icon is set in syncAttentionIcons
		QLabel *attn = new QLabel("");
		attn->setVisible(false);
		layout->addWidget(attn, row, col++, Qt::AlignLeft);
		attentionLabels.append(attn);
	}

	channel_id.append(chn);

	prev_iec = -1;
	recolorizeThermocouple();
	syncAttentionIcons();
}

DashSensorDevice::~DashSensorDevice()
{
	QLayoutItem *child;
	while ((child = layout->takeAt(0)) != 0) {
		delete child;
	}
	delete layout;
}

void DashSensorDevice::syncAttentionIcons(void)
{
	QSettings settings;
	USBTenki_channel *chn;
	int i, j;
	bool show_cal_buttons;

	show_cal_buttons = settings.value("ui/show_cal_buttons").toBool();

	for (i=0; i<attentionLabels.size(); i++) {

		chn = tenki_device->getChannelData(i);
		if (chn->num_calpoints > 0) {
			// If user calibration is disabled...
			if (!tenki_device->userCalibrationEnabled(i)) {
				// And any of the points is set, show a warning.
				for (j=0; j<chn->num_calpoints; j++) {
					if (usbtenki_isCalpointSet(&chn->calpoints[j])) {
						attentionLabels.at(i)->setText(LBL_CAL_SUSPENDED);
						attentionLabels.at(i)->setToolTip(TOOLTIP_CAL_SUSPENDED);
						attentionLabels.at(i)->setVisible(true);
						break;
					}
				}
				if (j >= chn->num_calpoints) {
					// If no calibration points are set, don't show a warning.
					attentionLabels.at(i)->setVisible(false);
				}
			} else {
				// User calibration is enabled. Show the enabled icon if there
				// are defined points.
				for (j=0; j<chn->num_calpoints; j++) {
					if (usbtenki_isCalpointSet(&chn->calpoints[j])) {
						attentionLabels.at(i)->setVisible(true);
						attentionLabels.at(i)->setToolTip(TOOLTIP_CAL_ACTIVE);
						attentionLabels.at(i)->setText(LBL_CAL_ENABLED);
						break;
					}
				}

				if (j >= chn->num_calpoints) {
					// Don't show a warning when user calibration is enabled and there
					// are no points.
					attentionLabels.at(i)->setVisible(false);
				}
			}
		}
	}

	for (i=0; i<calButtons.size(); i++) {
		if (show_cal_buttons) {
			calButtons.at(i)->setVisible(true);
		} else {
			calButtons.at(i)->setVisible(false);
		}
	}

}

void DashSensorDevice::recolorizeThermocouple(void)
{
	QString b;
	int use_iec;

	use_iec = g_tenkisources->getUseIECthermocoupleColors();

	if (use_iec == prev_iec) {
		return; // nothing to do
	}
	prev_iec = use_iec;

	for (int i=0; i<chip_ids.size(); i++)
	{
		b = QString::fromLocal8Bit(chipToString(chip_ids.at(i)));
		if (b.startsWith("Type-")) {

			if (use_iec) {
				b.replace("Type-K", "<font style='background-color: #40ab45; color: white;'><b>&nbsp;Type-K&nbsp;</b></font>");
				b.replace("Type-J", "<font style='background-color: #000000; color: white;'><b>&nbsp;Type-J&nbsp;</b></font>");
				b.replace("Type-N", "<font style='background-color: #fac9e1; color: black;'><b>&nbsp;Type-N&nbsp;</b></font>");
				b.replace("Type-E", "<font style='background-color: #6a3f92; color: white;'><b>&nbsp;Type-E&nbsp;</b></font>");
				b.replace("Type-T", "<font style='background-color: #794205; color: white;'><b>&nbsp;Type-T&nbsp;</b></font>");
				b.replace("Type-B", "<font style='background-color: #808080; color: white;'><b>&nbsp;Type-B&nbsp;</b></font>");
				b.replace("Type-S", "<font style='background-color: #e38538; color: white;'><b>&nbsp;Type-S&nbsp;</b></font>");
				b.replace("Type-R", "<font style='background-color: #e38538; color: white;'><b>&nbsp;Type-R&nbsp;</b></font>");
			} else {
				b.replace("Type-K", "<font style='background-color: #ffff00; color: black;'><b>&nbsp;Type-K&nbsp;</b></font>");
				b.replace("Type-J", "<font style='background-color: #000000; color: white;'><b>&nbsp;Type-J&nbsp;</b></font>");
				b.replace("Type-N", "<font style='background-color: #e2811f; color: white;'><b>&nbsp;Type-N&nbsp;</b></font>");
				b.replace("Type-E", "<font style='background-color: #d9018b; color: white;'><b>&nbsp;Type-E&nbsp;</b></font>");
				b.replace("Type-T", "<font style='background-color: #2350a0; color: white;'><b>&nbsp;Type-T&nbsp;</b></font>");
				b.replace("Type-B", "<font style='background-color: #808080; color: white;'><b>&nbsp;Type-B&nbsp;</b></font>");
				b.replace("Type-S", "<font style='background-color: #3f8d3e; color: white;'><b>&nbsp;Type-S&nbsp;</b></font>");
				b.replace("Type-R", "<font style='background-color: #3f8d3e; color: white;'><b>&nbsp;Type-R&nbsp;</b></font>");
			}
		}

		descriptions.at(i)->setText(b);
	}
}

void DashSensorDevice::refresh()
{
	double value;
	USBTenki_channel ch, *cdat;
//	qDebug() << "DashSensorDevice::refresh()";

	QString new_title = base_title;
	if (tenki_device->getStatus() == TENKI_DEVICE_STATUS_UNABLE_TO_OPEN) {
		new_title += " [ERROR]";
	}
	setTitle(new_title);

	for (int i=0; i<values.size(); i++) {
		QString d,e;

		// react to setting changes
		recolorizeThermocouple();

		cdat = tenki_device->getChannelData(channel_id.at(i));
		if (!cdat) {
			values.at(i)->setText("Error");
			continue;
		}
		if (cdat->status != USBTENKI_CHN_STATUS_VALID) {
			values.at(i)->setText(usbtenki_getChannelStatusString(cdat));
			continue;
		}

		g_tenkisources->convertToUnits(cdat, &ch);


		value = usbtenki_channelValue(&ch);

		g_tenkisources->formatValue(&d, value);
		if (chip_ids.at(i) == USBTENKI_CHIP_HEXCOLOR) {
			char tmpbuf[128];
			snprintf(tmpbuf, 128, "<font style='background-color: #%06x'; color: #%06x'><b>#%06x</b></font>",
				(uint32_t)value,
				0x000000,
				(uint32_t)value
				);
			values.at(i)->setText(tmpbuf);
		} else {
			values.at(i)->setText(d);
		}
		minimums.at(i)->submitValue(value);
		maximums.at(i)->submitValue(value);

		// those two QList are populated in the same
		// order. So it will be the same index i.
		e = QString::fromUtf8(unitToString(ch.converted_unit, 0));
		units.at(i)->setText(e);

	//	qDebug() << d;
	}

	syncAttentionIcons();
}

void DashSensorDevice::infoClicked(bool checked)
{
	QMessageBox msgBox;
	int major, minor;

	(void)checked;

	tenki_device->getFirmwareVersion(&major, &minor);

	QString text =	"<html><body><h1>" + QString(tenki_device->getProductName()) + "</h1>" +
					"Firmware version: " + QString::number(major) + "." + QString::number(minor) + "<br>" +
					"Serial number: " + QString(tenki_device->getSerial()) + "<br>" +
					"Channels: " + QString::number(tenki_device->getVisibleChannels()) +

					"</body></html>";

	msgBox.setWindowTitle("About sensor " + base_title );
	msgBox.setText(text);
	msgBox.setStandardButtons(QMessageBox::Ok);
	msgBox.setIcon(QMessageBox::Information);

	msgBox.exec();
}

void DashSensorDevice::cbAllBigviewChanged(int state)
{
	for (int i = 0; i<bigviewCheckboxes.count(); i++) {
		bigviewCheckboxes.at(i)->setChecked(state);
	}
}

void DashSensorDevice::cbAllGraphChanged(int state)
{
	for (int i = 0; i<graphCheckboxes.count(); i++) {
		graphCheckboxes.at(i)->setChecked(state);
	}
}

void DashSensorDevice::calibrateClicked(int id)
{
	CalibrationEditDialog *caldialog;

	caldialog = new CalibrationEditDialog(this, tenki_device, id);
	caldialog->exec();

	bool cal = caldialog->calibrationDisabled();
	tenki_device->enableUserCalibration(id, !cal);

	delete caldialog;

	//qDebug() << "Calibrate " + QString::number(id) + "Done.";
}

void DashSensorDevice::cbCalibrateChanged(int id, int checked)
{
	tenki_device->enableUserCalibration(id, checked);
}

